import os
import platform
import re
import shlex
import sys
import textwrap
from contextlib import contextmanager
from contextlib import nullcontext
from itertools import chain
from pathlib import Path
from textwrap import dedent
from typing import Any
from typing import Callable
from typing import ContextManager
from typing import Dict
from typing import Iterable
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Union
from warnings import warn

import jpype

from paquo._utils import QuPathVersion

# types
PathOrStr = Union[Path, str]

__all__ = ["JClass", "start_jvm", "find_qupath"]

JClass = jpype.JClass


class QuPathJVMInfo(NamedTuple):
    app_dir: Path
    runtime_dir: Path
    jvm_path: Path
    jvm_options: List[str]


def find_qupath(*,
                qupath_dir: Optional[PathOrStr] = None,
                qupath_search_dirs: Optional[Union[PathOrStr, List[PathOrStr]]] = None,
                qupath_search_dir_regex: Optional[str] = None,
                qupath_search_conda: Optional[bool] = None,
                qupath_prefer_conda: Optional[bool] = None,
                java_opts: Optional[Union[List[str], str]] = None,
                jvm_path_override: Optional[PathOrStr] = None,
                **_kwargs) -> QuPathJVMInfo:
    """find current qupath installation and jvm paths/options

    For now this supports a qupath which ships its own JRE installation
    and was installed via `conda install -c sdvillal qupath`

    Returns
    -------
    qupath_jvm_info:
        a tuple (app_dir, runtime_dir, jvm_path, jvm_options)
    """
    if java_opts is None:
        java_opts = []
    elif isinstance(java_opts, str):
        java_opts = shlex.split(java_opts)

    if jvm_path_override:
        jvm_path_override = Path(jvm_path_override)
        if not jvm_path_override.exists():
            raise FileNotFoundError(jvm_path_override)
        if jvm_path_override.is_dir():
            raise IsADirectoryError(jvm_path_override)
    else:
        jvm_path_override = None

    if qupath_dir:
        # short circuit in case we provide a qupath_dir
        # --> when qupath_dir is provided, search is disabled
        if isinstance(qupath_dir, str):
            qupath_dir = Path(qupath_dir)
        info = qupath_jvm_info_from_qupath_dir(qupath_dir, java_opts)
        if jvm_path_override:
            return info._replace(jvm_path=jvm_path_override)
        return info

    if qupath_search_dirs is None:
        qupath_search_dirs = []
    if not isinstance(qupath_search_dirs, list):
        qupath_search_dirs = [qupath_search_dirs]
    if qupath_search_dir_regex is None:
        qupath_search_dir_regex = ""  # match everything
    search_dirs = _scan_qupath_dirs(qupath_search_dirs, qupath_search_dir_regex)

    if qupath_search_conda:
        conda_search_dir = _conda_qupath_dir()
        if conda_search_dir is not None:
            if qupath_prefer_conda:
                search_dirs = chain([conda_search_dir], search_dirs)
            else:
                search_dirs = chain(search_dirs, [conda_search_dir])

    for qupath_dir in search_dirs:
        try:
            info = qupath_jvm_info_from_qupath_dir(qupath_dir, java_opts)
        except FileNotFoundError:
            continue
        else:
            if jvm_path_override:
                return info._replace(jvm_path=jvm_path_override)
            return info
    else:
        raise ValueError("no valid qupath installation found")


def _scan_qupath_dirs(qupath_search_dirs: List[PathOrStr], qupath_search_dir_regex: str) -> Iterable[Path]:
    """return potential paths for QuPath"""
    qp_match = re.compile(qupath_search_dir_regex).match
    for location in map(Path, qupath_search_dirs):
        if not location.is_dir():
            continue
        with os.scandir(location.absolute()) as it:
            for dir_entry in sorted(it, key=lambda x: x.name.lower(), reverse=True):
                if dir_entry.is_dir() and qp_match(dir_entry.name):
                    yield Path(dir_entry.path)


def _conda_qupath_dir() -> Optional[Path]:
    """return the conda qupath if running in a conda env"""
    prefix = os.environ.get('CONDA_PREFIX')
    if prefix:
        system = platform.system()
        if system == "Linux":
            return Path(prefix) / "opt" / "QuPath"
        elif system == "Darwin":
            return Path(prefix) / "bin" / "QuPath.app"
        elif system == "Windows":
            return Path(prefix) / "Library" / "QuPath"
        else:  # pragma: no cover
            raise ValueError(f'Unknown platform {system}')
    return None


def qupath_jvm_info_from_qupath_dir(qupath_dir: Path, jvm_options: List[str]) -> QuPathJVMInfo:
    """convert qupath_dir to paths according to platform"""
    system = platform.system()
    if system == "Linux":
        app_dir = qupath_dir / "lib" / "app"
        runtime_dir = qupath_dir / "lib" / "runtime"
        jvm_path = runtime_dir / "lib" / "server" / "libjvm.so"

    elif system == "Darwin":
        app_dir = qupath_dir / "Contents" / "app"
        runtime_dir = qupath_dir / "Contents" / "runtime" / "Contents" / "Home"
        jvm_path = runtime_dir / "lib" / "libjli.dylib"  # not server/libjvm.dylib

    elif system == "Windows":
        app_dir = qupath_dir / "app"
        runtime_dir = qupath_dir / "runtime"
        jvm_path = runtime_dir / "bin" / "server" / "jvm.dll"

    else:  # pragma: no cover
        raise ValueError(f'Unknown platform {system}')

    # verify that paths are sane
    if not (app_dir.is_dir() and runtime_dir.is_dir() and jvm_path.is_file()):
        raise FileNotFoundError('qupath installation is incompatible')

    # Add java.library.path so that the qupath provided openslide works
    jvm_options.append(f"-Djava.library.path={app_dir}")
    jvm_options = list(dict.fromkeys(jvm_options))  # keep options unique and in order

    return QuPathJVMInfo(app_dir, runtime_dir, jvm_path, jvm_options)


# stores qupath version to handle consecutive calls to start_jvm
_QUPATH_VERSION: Optional[QuPathVersion] = None


def start_jvm(
    finder: Optional[Callable[..., QuPathJVMInfo]] = None,
    finder_kwargs: Optional[Dict[str, Any]] = None,
) -> Optional[QuPathVersion]:
    """start the jvm via jpype

    This is automatically called at import of `paquo.java`.
    """
    global _QUPATH_VERSION

    if jpype.isJVMStarted():
        return _QUPATH_VERSION  # nothing to be done

    if finder is None:
        finder = find_qupath

    if finder_kwargs is None:
        finder_kwargs = {}  # pragma: no cover

    # For the time being, we assume qupath is our JVM of choice
    app_dir, runtime_dir, jvm_path, jvm_options = finder(**finder_kwargs)

    patched_env: Callable[[], ContextManager[Any]]
    if platform.system() == "Windows":
        # workaround for EXCEPTION_ACCESS_VIOLATION crash
        # see: https://github.com/bayer-science-for-a-better-life/paquo/issues/67
        @contextmanager
        def patched_env():
            _old = os.environ.copy()
            os.environ.update({
                "PATH": f"{os.path.join(runtime_dir, 'bin')}{os.pathsep}{os.environ['PATH']}"
            })
            try:
                yield
            finally:
                os.environ.clear()
                os.environ.update(_old)

        # the above workaround doesn't fix the issue for python versions installed
        # via the Microsoft Store. Let's warn users that this might cause problems
        def is_windows_store_python() -> bool:
            parts = Path(sys.base_exec_prefix).parts
            try:
                idx = parts.index("WindowsApps")
            except ValueError:
                return False
            try:
                return parts[idx + 1].startswith("PythonSoftwareFoundation")
            except IndexError:
                return False

        if finder_kwargs.pop("warn_microsoft_store_python", True) and is_windows_store_python():
            msg = dedent("""\
            MicrosoftStore Python installation detected
            Your Python version seems to be installed via the MicrosoftStore.
            If paquo crashes with a EXCEPTION_ACCESS_VIOLATION try installing Python from https://www.python.org
            To silence this warning set the following in your .paquo.toml configfile:
            >>> warn_microsoft_store_python = false <<<
            """)
            warn(msg, stacklevel=2)

    else:
        patched_env = nullcontext

    # This is not really needed, but beware we might need SL4J classes (see warning)
    jpype.addClassPath(str(app_dir / '*'))

    with patched_env():
        try:
            jpype.startJVM(
                str(jvm_path),
                *jvm_options,
                ignoreUnrecognized=False,
                convertStrings=False
            )
        except OSError as err:
            if (
                err.errno == 0
                and "JVM DLL not found:" in str(err)
                and jvm_path.is_file()
                and platform.uname().machine == "arm64"
            ):
                msg = textwrap.dedent("""\
                Probably a JVM and Python architecture issue on M1:
                You can fix this by running a JVM with the same architecture as your
                Python interpreter. Usually paquo uses the JVM that ships with QuPath,
                but you can override this by setting 'jvm_path_override' in `.paquo.toml`

                In case you are running an arm64 Python interpreter on your mac try
                installing an arm jvm from
                https://www.azul.com/downloads/?version=java-17-lts&os=macos&architecture=arm-64-bit&package=jdk
                ```
                $ cat .paquo.toml
                # use the path to libjli.dylib valid on your machine:
                jvm_path_override = "/Library/Java/JavaVirtualMachines/zulu-17.jdk/Contents/Home/lib/libjli.dylib"
                ```
                If this doesn't work, please open an issue on github.
                """)
                raise RuntimeError(msg) from err
        except RuntimeError as jvm_error:  # pragma: no cover
            # there's a chance that this RuntimeError occurred because a user provided
            # jvm_option is incorrect. let's try if that is the case and crash with a
            # more verbose error message
            try:
                jpype.startJVM(
                    str(jvm_path),
                    *jvm_options,
                    ignoreUnrecognized=True,
                    convertStrings=False
                )
            except RuntimeError:
                raise jvm_error
            else:
                msg = f"Provided JAVA_OPTS prevent the JVM from starting! {jvm_options}"
                exc = RuntimeError(msg)
                exc.__cause__ = jvm_error
                raise exc

    # we'll do this explicitly here to verify the QuPath version
    try:
        _version = str(JClass("qupath.lib.common.GeneralTools").getVersion())
    except (TypeError, AttributeError):  # pragma: no cover
        version = _QUPATH_VERSION = None
    else:
        version = _QUPATH_VERSION = QuPathVersion(_version)
    return version



import warnings

from paquo._config import settings
from paquo._config import to_kwargs
from paquo._utils import QuPathVersion
from paquo.jpype_backend import JClass
from paquo.jpype_backend import start_jvm

# we can extend this as when we add more testing against different versions
MIN_QUPATH_VERSION = QuPathVersion('0.2.0')  # FIXME: this is bound to change


# allow paquo to be imported in case qupath and a jvm are not available
# Note: this renders paquo unusable. But we need it for example for the
#   sphinx docs to be generated without requiring an installed qupath.
if settings.mock_backend:  # pragma: no cover
    from unittest.mock import MagicMock
    from unittest.mock import create_autospec

    start_jvm = create_autospec(start_jvm, return_value=MIN_QUPATH_VERSION)
    # noinspection PyPep8Naming
    def JClass(jc, *_args, **_kwargs):  # noqa
        class _JClassType(type):
            def __getattr__(cls, key):
                return MagicMock()

        class _JClass(metaclass=_JClassType):
            f"""Java Class: {jc!r}"""
        return _JClass

# ensure the jvm is running
qupath_version = start_jvm(finder_kwargs=to_kwargs(settings))
if qupath_version is None:
    # let's not exit for now but warn the user:
    warnings.warn("COULD NOT DETECT QUPATH VERSION! UNSUPPORTED")  # pragma: no cover
elif qupath_version < MIN_QUPATH_VERSION:
    # let's not exit for now but warn the user:
    warnings.warn(f"QUPATH '{qupath_version}' IS UNTESTED OR UNSUPPORTED")  # pragma: no cover


class _Compatibility:
    """organizes QuPath version differences"""
    def __init__(self, version: "QuPathVersion | None") -> None:
        self.version = version

    def requires_missing_classes_json_fix(self) -> bool:
        # older QuPaths crash on project load when classes.json is missing
        # see: https://github.com/qupath/qupath/commit/be861cea80b9a8ef300e30d7985fd69791c2432e
        if self.version is None:
            return True
        else:
            return self.version <= QuPathVersion("0.2.0")

    def requires_annotation_json_fix(self) -> bool:
        # annotations changed between QuPath "0.2.3" and "0.3.x"
        # see: https://github.com/qupath/qupath/commit/fef5c43ce3f67e0e062677c407b395ef3e6e27c3
        if self.version is None:
            return True
        else:
            return self.version <= QuPathVersion("0.2.3")

    def supports_image_server_recovery(self) -> bool:
        # image_server server.json files are only guaranteed to be written since QuPath "0.2.0"
        # see: https://github.com/qupath/qupath/commit/39abee3012da9252ea988308848c5d802164e060
        if self.version is None:
            return False
        else:
            return self.version >= QuPathVersion("0.2.0")

    def supports_logmanager(self) -> bool:
        # the logmanager class was only added with 0.2.0-m10
        # see: https://github.com/qupath/qupath/commit/15b844703b686f7a9a64c50194ebe22fc46924a5
        if self.version is None:
            return False
        else:
            return self.version >= QuPathVersion("0.2.0-m10")

    def supports_newer_addobject_and_pathclass(self) -> bool:
        # PathObjectHierarchy.addPathObject and .addPathObjectWithoutUpdate are deprecated
        # PathClassFactory is deprecated too
        # see: https://github.com/qupath/qupath/commit/b961944120c26a3a4cfe01959d8e8b993075ffdf
        if self.version is None:
            return False
        else:
            return self.version >= QuPathVersion("0.4.0")


compatibility = _Compatibility(qupath_version)


ArrayList = JClass("java.util.ArrayList")
BufferedImage = JClass('java.awt.image.BufferedImage')
ByteArrayOutputStream = JClass("java.io.ByteArrayOutputStream")
File = JClass('java.io.File')
Files = JClass('java.nio.file.Files')
Integer = JClass('java.lang.Integer')
PrintStream = JClass('java.io.PrintStream')
StandardCharsets = JClass("java.nio.charset.StandardCharsets")
String = JClass('java.lang.String')
System = JClass('java.lang.System')
URI = JClass('java.net.URI')

ColorTools = JClass('qupath.lib.common.ColorTools')
DefaultProject = JClass('qupath.lib.projects.DefaultProject')
DefaultProjectImageEntry = JClass('qupath.lib.projects.DefaultProject.DefaultProjectImageEntry')
GeneralTools = JClass("qupath.lib.common.GeneralTools")
GeometryTools = JClass("qupath.lib.roi.GeometryTools")
GsonTools = JClass('qupath.lib.io.GsonTools')
ImageData = JClass('qupath.lib.images.ImageData')
ImageType = JClass('qupath.lib.images.ImageData.ImageType')
ImageServer = JClass('qupath.lib.images.servers.ImageServer')
ImageServers = JClass('qupath.lib.images.servers.ImageServers')  # NOTE: this is needed to make QuPath v0.3.0-rc1 work
ImageServerProvider = JClass('qupath.lib.images.servers.ImageServerProvider')

if compatibility.supports_logmanager():
    LogManager = JClass('qupath.lib.gui.logging.LogManager')
else:
    LogManager = None

PathAnnotationObject = JClass("qupath.lib.objects.PathAnnotationObject")
PathClass = JClass('qupath.lib.objects.classes.PathClass')

if not compatibility.supports_newer_addobject_and_pathclass():
    PathClassFactory = JClass('qupath.lib.objects.classes.PathClassFactory')
else:
    PathClassFactory = None

PathDetectionObject = JClass("qupath.lib.objects.PathDetectionObject")
PathIO = JClass("qupath.lib.io.PathIO")
PathObjectHierarchy = JClass('qupath.lib.objects.hierarchy.PathObjectHierarchy')
PathObjects = JClass("qupath.lib.objects.PathObjects")
PathROIObject = JClass("qupath.lib.objects.PathROIObject")
PathTileObject = JClass("qupath.lib.objects.PathTileObject")
PathCellObject = JClass("qupath.lib.objects.PathCellObject")
Point2 = JClass("qupath.lib.geom.Point2")
ProjectIO = JClass('qupath.lib.projects.ProjectIO')
Projects = JClass('qupath.lib.projects.Projects')
ROI = JClass("qupath.lib.roi.interfaces.ROI")
ROIs = JClass("qupath.lib.roi.ROIs")
ServerTools = JClass('qupath.lib.images.servers.ServerTools')

EllipseROI = JClass("qupath.lib.roi.EllipseROI")
GeometryROI = JClass("qupath.lib.roi.GeometryROI")
LineROI = JClass("qupath.lib.roi.LineROI")
PointsROI = JClass("qupath.lib.roi.PointsROI")
PolygonROI = JClass("qupath.lib.roi.PolygonROI")
PolylineROI = JClass("qupath.lib.roi.PolylineROI")
RectangleROI = JClass("qupath.lib.roi.RectangleROI")

WKBWriter = JClass("org.locationtech.jts.io.WKBWriter")
WKBReader = JClass("org.locationtech.jts.io.WKBReader")

IOException = JClass("java.io.IOException")
ExceptionInInitializerError = JClass("java.lang.ExceptionInInitializerError")
URISyntaxException = JClass("java.net.URISyntaxException")
NegativeArraySizeException = JClass('java.lang.NegativeArraySizeException')
IllegalArgumentException = JClass('java.lang.IllegalArgumentException')
FileNotFoundException = JClass('java.io.FileNotFoundException')
NoSuchFileException = JClass('java.nio.file.NoSuchFileException')


def __getattr__(name):
    """lazy import some"""
    if name == "ProjectImportImagesCommand":
        warnings.warn(
            "ProjectImportImagesCommand will be removed from paquo.java",
            DeprecationWarning
        )
        return JClass('qupath.lib.gui.commands.ProjectImportImagesCommand')
    else:
        raise AttributeError(name)


# noinspection PyPep8Naming
def ProjectImportImagesCommand_getThumbnailRGB(server, _):
    # needs to be lazily imported to not emit threading info message
    pcmd = JClass('qupath.lib.gui.commands.ProjectImportImagesCommand')
    return pcmd.getThumbnailRGB(server, None)



import os
from unittest import mock

from paquo._config import _get_settings
from paquo._config import to_kwargs


def test_double_underscore_envvar_override():
    with mock.patch.dict(os.environ, {"PAQUO__QUPATH_DIR": "/somewhere"}):
        settings = _get_settings()
        kw = to_kwargs(settings)
    assert kw["qupath_dir"] == "/somewhere"
    assert "_qupath_dir" not in kw


def test_single_underscore_envvar_override():
    with mock.patch.dict(os.environ, {"PAQUO_QUPATH_DIR": "/somewhere"}):
        settings = _get_settings()
        kw = to_kwargs(settings)
    assert kw["qupath_dir"] == "/somewhere"



import hashlib
import pathlib
import shutil
import sys
import urllib.request

import pytest

# openslide aperio test images
IMAGES_BASE_URL = "http://openslide.cs.cmu.edu/download/openslide-testdata/Aperio/"


def md5(fn):
    if sys.version_info >= (3, 9):
        m = hashlib.md5(usedforsecurity=False)
    else:
        m = hashlib.md5()  # nosec B324
    with open(fn, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            m.update(chunk)
    return m.hexdigest()


@pytest.fixture(scope='session')
def svs_small():
    """download the smallest aperio test image svs"""
    small_image = "CMU-1-Small-Region.svs"
    small_image_md5 = "1ad6e35c9d17e4d85fb7e3143b328efe"
    data_dir = pathlib.Path(__file__).parent / "data"

    data_dir.mkdir(parents=True, exist_ok=True)
    img_fn = data_dir / small_image

    if not img_fn.is_file():
        # download svs from openslide test images
        url = IMAGES_BASE_URL + small_image
        with urllib.request.urlopen(url) as response, open(img_fn, 'wb') as out_file:  # nosec B310
            shutil.copyfileobj(response, out_file)

    if md5(img_fn) != small_image_md5:  # pragma: no cover
        shutil.rmtree(img_fn)
        pytest.fail("incorrect md5")
    else:
        yield img_fn.absolute()


@pytest.fixture(scope="function")
def renamed_svs_small(request, tmp_path, svs_small):
    name = request.param
    new_path = tmp_path.joinpath(name)
    new_path.write_bytes(svs_small.read_bytes())
    yield new_path


@pytest.fixture(scope='session')
def qupath_version():
    from paquo.java import qupath_version
    yield qupath_version



import os
import platform
import posixpath
import shutil
import tempfile
from contextlib import nullcontext
from pathlib import Path

import pytest

from paquo._utils import QuPathVersion
from paquo.images import ImageProvider, QuPathProjectImageEntry, QuPathImageType
from paquo.projects import QuPathProject


@pytest.fixture(scope='function')
def new_project(tmp_path):
    yield QuPathProject(tmp_path / "paquo-project", mode='x')


@pytest.fixture(scope='function')
def tmp_path2(tmp_path):
    yield tmp_path / "another_path"


def test_project_instance():
    with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
        q = QuPathProject(tmpdir, mode='x')
        repr(q)
        q.save()


def test_project_create_no_dir():
    with tempfile.TemporaryDirectory(prefix='paquo-') as tmpdir:
        project_path = Path(tmpdir) / "new_project"
        q = QuPathProject(project_path, mode='x')
        q.save()


def test_project_creation_input_error(tmp_path):

    p = tmp_path / Path('somewhere')
    p.mkdir()

    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        QuPathProject(p, image_provider=object())

    with pytest.raises(ValueError):
        QuPathProject(p / 'myproject.proj')  # needs .qpproj

    with pytest.raises(FileNotFoundError):
        QuPathProject(p / 'myproject.qpproj', mode='r+')

    with open(p / 'should-not-be-here', 'w') as f:
        f.write('project directories need to be empty')
    with pytest.raises(ValueError):
        QuPathProject(p, mode='x')


# noinspection PyTypeChecker
@pytest.mark.parametrize(
    "mode", [
        'r', 'r+', 'w', 'w+', 'a', 'a+', 'x', 'x+'
    ]
)
def test_project_creation_mode(mode, tmp_path2, new_project):
    # test creating in empty dir
    cm = pytest.raises(FileNotFoundError) if "r" in mode else nullcontext()
    with cm:
        QuPathProject(tmp_path2, mode=mode).save()

    # prepare an existing proj
    p = new_project.path
    new_project.save()
    del new_project
    assert not p.with_suffix('.qpproj.backup').is_file()
    assert not any(p.parent.parent.glob('*.backup'))

    # test creating in existing dir
    if "x" in mode:
        cm = pytest.raises(FileExistsError)
    elif "r" == mode:
        cm = pytest.raises(IOError)  # can't save!
    else:
        cm = nullcontext()
    with cm:
        QuPathProject(p, mode=mode).save()

    assert p.is_file()
    backups = list(p.parent.parent.glob('*.backup'))

    if 'w' in mode:
        assert len(backups) == 1
        assert backups[0].is_file()
    else:
        assert len(backups) == 0


def test_backup_path_project_stash(tmp_path):
    from paquo.projects import _stash_project_files
    p = Path(tmp_path)
    assert _stash_project_files(p) is None
    f = p / "wrong.file"
    f.touch()
    assert _stash_project_files(f) is None


# noinspection PyTypeChecker
def test_unsupported_mode(tmp_path):
    with pytest.raises(ValueError):
        QuPathProject(tmp_path, mode="???")


def test_project_open_with_filename(new_project):
    new_project.save()
    # this points to path/project.qpproj
    proj_fn = new_project.path
    QuPathProject(proj_fn, mode='r+')


def test_project_name(tmp_path):
    p = Path(tmp_path) / "my_project_123"
    p.mkdir()
    qp = QuPathProject(p, mode='x')
    assert qp.name == "my_project_123"


def test_project_uri(new_project):
    assert new_project.uri.startswith("file:")
    assert new_project.uri.endswith(".qpproj")
    # uri_previous is None for empty projects
    # assert new_project.uri_previous is None


def test_project_save_and_path(new_project):
    assert not new_project.path.is_file()
    new_project.save()
    assert new_project.path.is_file()


def test_project_version(new_project):
    from paquo.java import GeneralTools
    new_project.save()
    qp_version = str(GeneralTools.getVersion())
    assert new_project.version == qp_version

    with QuPathProject(new_project.path, mode='r+') as qp:
        assert qp.version == qp_version


def test_project_add_path_classes(new_project):
    from paquo.classes import QuPathPathClass

    names = {'a', 'b', 'c'}
    new_project.path_classes = map(QuPathPathClass, names)

    assert len(new_project.path_classes) == 3
    assert {c.name for c in new_project.path_classes} == names


def test_download_svs(svs_small):
    assert svs_small.is_file()


def test_timestamps(new_project):
    assert new_project.timestamp_creation > 0
    assert new_project.timestamp_modification > 0


def test_project_add_image(new_project, svs_small):
    with pytest.raises(ValueError):
        # can't add images by instantiating them
        QuPathProjectImageEntry(svs_small)

    entry = new_project.add_image(svs_small)
    assert (Path(entry.entry_path) / "thumbnail.jpg").is_file()

    assert len(new_project.images) == 1
    assert entry in new_project.images
    assert object() not in new_project.images


@pytest.mark.skipif(platform.system() != "Windows", reason="windows only")
def test_project_add_image_windows(new_project, svs_small):
    drive, *parts = svs_small.parts

    uri = posixpath.join(f"file:////localhost/{drive[0].lower()}$", *parts)

    _ = new_project.add_image(uri)
    assert len(new_project.images) == 1


@pytest.mark.skipif(platform.system() != "Windows", reason="windows only")
@pytest.mark.parametrize(
    "renamed_svs_small", [
        pytest.param("abc.svs", id="simple_name"),
        pytest.param("image%image.svs", id="name_with_percentage"),
    ],
    indirect=True,
)
def test_project_add_image_windows_unencoded_uri(new_project, renamed_svs_small):
    drive, *parts = renamed_svs_small.parts

    uri = posixpath.join(f"file:////localhost/{drive[0]}$", *parts)

    _ = new_project.add_image(uri)
    assert len(new_project.images) == 1


def test_project_add_image_writes_project(tmp_path, svs_small):
    qp = QuPathProject(tmp_path, mode='x')
    qp.add_image(svs_small)

    assert qp.path.is_file()


def test_project_add_image_twice(new_project, svs_small):
    new_project.add_image(svs_small)
    with pytest.raises(FileExistsError):
        new_project.add_image(svs_small)
    assert len(new_project.images) == 1

    new_project.add_image(svs_small, allow_duplicates=True)
    assert len(new_project.images) == 2


def test_project_add_image_with_type(new_project, svs_small):
    t = QuPathImageType.BRIGHTFIELD_H_DAB
    entry = new_project.add_image(svs_small, image_type=t)
    assert entry.image_type == t


def test_project_add_unsupported_image(new_project, tmp_path):
    image = Path(tmp_path) / "unsupported.image"
    with open(image, "w") as f:
        f.write("very unsupported image")

    with pytest.raises(IOError):
        new_project.add_image(image)


def test_project_remove_image(new_project, svs_small):
    # Add an image to the project and then remove it:
    new_project.add_image(svs_small)
    assert len(new_project.images) == 1
    new_project.remove_image(new_project.images[0])
    assert len(new_project.images) == 0


def test_project_remove_image_wrong_type(new_project):
    with pytest.raises(TypeError):
        new_project.remove_image("some/file.svs")


def test_project_image_slicing(new_project):
    _ = new_project.images[slice(None, None, None)]


def test_project_image_incorrect_index(new_project):
    with pytest.raises(IndexError):
        # noinspection PyTypeChecker
        _ = new_project.images["abc"]


def test_project_image_repr(new_project):
    assert repr(new_project.images)


def test_project_add_image_incorrect_path(new_project):
    with pytest.raises(FileNotFoundError):
        new_project.add_image("i-do-not-exist.svs")


def test_project_save_image_data(new_project, svs_small):
    from paquo.pathobjects import QuPathPathAnnotationObject
    from shapely.geometry import Point
    entry = new_project.add_image(svs_small)
    entry.hierarchy.annotations.add(
        QuPathPathAnnotationObject.from_shapely(
            Point(1, 2)
        )
    )
    new_project.save()
    assert (entry.entry_path / "data.qpdata").is_file()


def test_project_delete_image_file_when_opened(new_project, svs_small, qupath_version):
    # prepare new image to be deleted
    new_svs_small = new_project.path.parent / f"image_be_gone{svs_small.suffix}"
    shutil.copy(svs_small, new_svs_small)

    entry = new_project.add_image(new_svs_small)
    assert entry.is_readable()

    # TODO: this test should actually depend on the backend that qupath
    #   uses internally for opening the file. Since we reenabled openslide
    #   we hardcode it here, so that the intention of this test stays clear
    qupath_uses = "OPENSLIDE"

    if qupath_uses == "BIOFORMATS":  # pragma: no cover
        if platform.system() == "Windows":
            # NOTE: on Windows because you can't delete files that have open
            #   file handles. In this test we're deleting the file opened by
            #   the ImageServer on the java side. (this is happening
            #   implicitly when calling is_readable() because the java
            #   implementation of is_readable() loads the ImageData which
            #   creates an instance of an ImageServer)
            cm = pytest.raises(PermissionError)
        else:
            cm = nullcontext()

        with cm:
            os.unlink(new_svs_small)

    elif qupath_uses == "OPENSLIDE":

        if (
            qupath_version >= QuPathVersion("0.5.0")
            and platform.system() == "Windows"
        ):
            cm = pytest.raises(PermissionError)
        else:
            cm = nullcontext()

        with cm:
            os.unlink(new_svs_small)

    else:  # pragma: no cover
        raise ValueError('...')


@pytest.mark.xfail(platform.system() == "Windows", reason="file handles don't get closed by qupath?")
def test_project_image_uri_update(tmp_path, svs_small):

    project_path = tmp_path / "paquo-project"
    new_svs_small = tmp_path / "images" / f"image_be_gone{svs_small.suffix}"
    new_svs_small.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(svs_small, new_svs_small)
    assert new_svs_small.is_file()

    # create a project
    with QuPathProject(project_path, mode='x') as qp:
        entry = qp.add_image(new_svs_small)
        assert entry.is_readable()
        assert all(qp.is_readable().values())

    # cleanup
    del entry
    del qp

    # remove image
    os.unlink(new_svs_small)

    # reload the project
    with QuPathProject(project_path, mode='r+') as qp:
        entry, = qp.images
        assert not entry.is_readable()
        assert not all(qp.is_readable().values())

        # test if there's no remapping
        qp.update_image_paths(uri2uri={})
        # test if we remap to same
        qp.update_image_paths(uri2uri={entry.uri: entry.uri})

        # create mapping for uris
        uri2uri = {
            entry.uri: ImageProvider.uri_from_path(svs_small),
        }
        # update the uris
        qp.update_image_paths(uri2uri=uri2uri)

        # test that entry can be read
        assert entry.is_readable()
        assert all(qp.is_readable().values())


def test_project_image_uri_update_try_relative(tmp_path, svs_small):

    # prepare initial location
    location_0 = tmp_path / "location_0"
    location_0.mkdir()

    # prepare a project and images at location_0
    image_path = location_0 / "images" / "image0.svs"
    image_path.parent.mkdir(parents=True)
    shutil.copy(svs_small, image_path)
    assert image_path.is_file()

    # create a project
    with QuPathProject(location_0 / "project", mode='x') as qp:
        entry = qp.add_image(image_path)
        assert entry.is_readable()
        assert all(qp.is_readable().values())

    # cleanup
    del entry
    del qp

    # NOW move the location
    location_1 = tmp_path / "location_1"
    try:
        shutil.move(location_0, location_1)
    except PermissionError:
        if platform.system() == "Windows":
            pytest.xfail("Windows QuPath==0.5.0 quirk with permissions")
        else:
            raise

    with QuPathProject(location_1 / "project", mode='r') as qp:
        entry, = qp.images
        assert not entry.is_readable()
        assert not all(qp.is_readable().values())
        assert 'location_1' in str(qp.path)
        assert 'location_0' in entry.uri  # still points to the old location

        # fixme: testing private api
        assert 'location_0' in str(qp.java_object.getPreviousURI().toString())

        qp.update_image_paths(try_relative=True)

        # test that entry can be read
        assert entry.is_readable()
        assert all(qp.is_readable().values())



from __future__ import annotations

import argparse
import functools
import os
import platform
import sys
import tempfile
from contextlib import redirect_stdout
from itertools import repeat
from logging.config import dictConfig
from pathlib import Path
from typing import Callable

from paquo._cli import DirectoryType
from paquo._cli import argument
from paquo._cli import config_print_defaults
from paquo._cli import config_print_settings
from paquo._cli import create_project
from paquo._cli import export_annotations
from paquo._cli import list_project
from paquo._cli import open_qupath
from paquo._cli import qpzip_project
from paquo._cli import subcommand

parser = argparse.ArgumentParser(
    prog="python -m paquo" if Path(sys.argv[0]).name == "__main__.py" else None,
    description="""\
 ██████╗  █████╗  ██████╗ ██╗   ██╗ ██████╗
 ██╔══██╗██╔══██╗██╔═══██╗██║   ██║██╔═══██╗
 ██████╔╝███████║██║   ██║██║   ██║██║   ██║
 ██╔═══╝ ██╔══██║██║▄▄ ██║██║   ██║██║   ██║
 ██║     ██║  ██║╚██████╔╝╚██████╔╝╚██████╔╝
 ╚═╝     ╚═╝  ╚═╝ ╚══▀▀═╝  ╚═════╝  ╚═════╝ """,
    epilog="#### [PA]thological [QU]path [O]bsession ####",
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
subparsers = parser.add_subparsers(dest="cmd", title="paquo command")
subcommand = functools.partial(subcommand, parent=subparsers)
parser.add_argument('--version', action='store_true', help="print paquo version")
parser.add_argument('--qupath-version', action='store_true', help="print qupath version")


def main(commandline=None):
    """main command line argument handling"""
    try:
        args = parser.parse_args(commandline)
    except UnicodeEncodeError:
        # recover in case terminal does not support unicode
        parser.description = "PAQUO"
        args = parser.parse_args(commandline)

    if args.cmd is None:
        if args.version or args.qupath_version:
            if args.version:
                from paquo import __version__
                print(f"{__version__}")
            if args.qupath_version:
                from paquo.java import qupath_version
                print(f"{qupath_version!s}")
            return 0
        else:
            parser.print_help()
            return 1
    else:
        from paquo._config import settings
        lvl = 'INFO'
        if settings.cli_force_log_level_error:
            lvl = 'ERROR'
        dictConfig({
            'version': 1,
            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                },
            },
            "loggers": {
                'paquo': {
                    'level': lvl,
                    'handlers': ['console'],
                },
                'qupath': {
                    'level': 'CRITICAL',
                    'handlers': ['console'],
                },
            },
        })
        return args.cmd_func(args)


@subcommand(
    argument('-l', '--list', action='store_true', help="list the paquo config"),
    argument('--default', action='store_true', help="default instead of current config"),
    argument(
        '-o', '--output',
        action='store',
        type=DirectoryType(), dest='output',
        help="directory where configuration is written to"
    ),
    argument('--force', action='store_true', help="force overwrite existing config"),
    argument('--search-tree', action='store_true', help="list all locations searched for config"),
)
def config(args, subparser):
    """handle paquo configuration"""
    from paquo._config import PAQUO_CONFIG_FILENAME
    from paquo._config import get_searchtree

    if not (args.list or args.search_tree):
        print(subparser.format_help())
        return 0

    if args.search_tree:
        print(f"paquo is scanning these dirs for '{PAQUO_CONFIG_FILENAME}':")
        for idx, location in enumerate(get_searchtree()):
            print(f"{idx}.", location)
        return 0

    if args.default:
        config_print = config_print_defaults
    else:
        config_print = config_print_settings

    if args.output is None:
        config_print()
    else:
        out_fn = args.output / PAQUO_CONFIG_FILENAME
        mode = "x" if not args.force else "w"
        # write to file
        try:
            with out_fn.open(mode) as f:
                with redirect_stdout(f):
                    config_print()
        except FileExistsError:
            print(f"ERROR: file {out_fn} exists! use --force to overwrite")
            return 1
    return 0


@subcommand(
    argument('project_path', nargs='?', default=None, help="path to your qupath project file/folder"),
)
def list_(args, subparser):
    """list contents of a qupath project"""
    if not args.project_path:
        print(subparser.format_help())
        return 0
    try:
        list_project(args.project_path)
    except FileNotFoundError as err:
        print(str(err), file=sys.stderr)
        return 1
    return 0


@subcommand(
    argument('project_path', nargs='?', default=None, help="path to new qupath project"),
    argument('--classes', nargs='*', help="path classes to add to project"),
    argument('--class-colors', nargs='*', help="path class colors in order of classes"),
    argument('--remove-default-classes', action="store_true", help="don't add default classes"),
    argument('--images', nargs='*', help="images to add to project"),
    argument('--force', action="store_true", help="force overwrite existing projects"),
)
def create(args, subparser):
    """create a new qupath project"""
    if not args.project_path:
        print(subparser.format_help())
        return 0

    if args.class_colors and len(args.class_colors) != len(args.classes):
        print("ERROR: need to specify same number of colors as classes")
        return 1

    if args.classes and len(args.classes) != len(set(args.classes)):
        print("ERROR: classes have to be unique")
        return 1

    class_names_colors = zip(args.classes or [], args.class_colors or repeat(None))

    images = args.images or []
    for image in images:
        if not Path(image).is_file():
            print(f"ERROR: image {image} is not a file")
            return 1

    try:
        name = create_project(
            args.project_path,
            class_names_colors=class_names_colors,
            images=images,
            remove_default_classes=args.remove_default_classes,
            force_write=args.force
        )
    except FileExistsError:
        print(f"ERROR: project {args.project_path} exists! use --force to overwrite")
        return 1
    print(f"project '{name}' created")
    return 0


@subcommand(
    argument('project_path', nargs='?', default=None, help="path to your qupath project file/folder"),
    argument('--image-idx', '-i', default=None, type=int, help="index of a qupath image"),
    argument(
        '-o', '--output',
        action='store',
        type=argparse.FileType('w'), dest='output',
        help="directory where configuration is written to"
    ),
    argument('--pretty', action='store_true', help="pretty format the output"),
)
def export(args, subparser):
    """export annotations of a qupath project image to geojson"""
    if args.project_path is None or args.image_idx is None:
        print(subparser.format_help())
        return 0

    try:
        if args.output is None:
            export_annotations(args.project_path, args.image_idx, args.pretty)
        else:
            with redirect_stdout(args.output):
                export_annotations(args.project_path, args.image_idx, args.pretty)
    except IndexError:
        print(f"ERROR: image index {args.image_idx} out of range")
        return 1
    return 0


@subcommand(
    argument('project_path', nargs='?', default=None, help="path to your qupath project file/folder"),
)
def open_(args, subparser):
    """open qupath with a specified project"""
    if args.project_path is None:
        print(subparser.format_help())
        return 0

    open_qupath(args.project_path)
    return 0


@subcommand(
    argument('project_path', nargs='?', default=None, help="path to your qupath project file/folder"),
)
def qpzip(args, subparser):
    """create a qpzip archive of a project"""
    if args.project_path is None:
        print(subparser.format_help())
        return 0

    qpzip_project(args.project_path)
    return 0


@subcommand(
    argument('image', nargs='?', default=None, help="path to image"),
    argument('--annotations', action="append", help="annotations for image"),
    argument('--annotations-cmd', type=str, help="automatically choose annotations if available")
)
def quickview(args, subparser):
    """open an image in qupath"""
    import subprocess

    if not args.image:
        print(subparser.format_help())
        return 0

    image = Path(args.image)
    if not image.is_file():
        print(f"ERROR: image {args.image} is not a file")
        return 1

    f_annotations = None
    if args.annotations:
        def f_annotations(name):  # noqa
            if name != image.name:
                return []
            return list(args.annotations)

    f_annotations_cmd = None
    if args.annotations_cmd:
        def f_annotations_cmd(name):  # noqa
            import shlex
            _cmd = shlex.split(f"{args.annotations_cmd} {name}")
            print("annotations", _cmd)
            output = subprocess.run(_cmd, env=os.environ, check=True, capture_output=True)
            return [line.decode() for line in output.stdout.splitlines() if line.strip()]

    cmd = None
    if f_annotations_cmd or f_annotations:
        def cmd(name):  # noqa
            names = []
            if f_annotations:
                names.extend(f_annotations(name))
            if f_annotations_cmd:
                names.extend(f_annotations_cmd(name))
            return names

    with tempfile.TemporaryDirectory() as project_path:
        create_project(project_path, class_names_colors=[], images=[image], annotations_json_func=cmd)
        open_qupath(project_path)

    return 0


@subcommand(
    argument("--install-path", type=DirectoryType(), required=True, help="extract / install QuPath to this path"),
    argument("--override-system", default=platform.system(), help=argparse.SUPPRESS),
    argument(
        "--download-path",
        type=DirectoryType(),
        default=tempfile.gettempdir(),
        help="download QuPath archives to this path",
    ),
    argument("--no-ssl-verify", action="store_true", help="(danger!) disable ssl verification for download"),
    argument("version", type=str, help="the QuPath version to download"),
)
def get_qupath(args, subparser):
    """download a specific QuPath version"""
    from paquo._utils import download_qupath
    from paquo._utils import extract_qupath

    system = args.override_system

    def _download_cb(it, name):
        if name:
            print("# downloading:", name)
        print("# progress ", end="", flush=True)
        try:
            for chunk in it:
                print(".", end="", flush=True)
                yield chunk
            print(" OK", end="", flush=True)
        finally:
            print("")

    try:
        file = download_qupath(
            args.version,
            path=args.download_path,
            callback=_download_cb,
            system=system,
            ssl_verify=not args.no_ssl_verify
        )
    except Exception:
        raise SystemExit(1)
    print("# extracting:", file)
    app = extract_qupath(file, args.install_path, system=system)
    print("# available at:", app)

    print("#\n# use via environment variable:")
    if system in {"Linux", "Darwin"}:
        print(f"#  $ export PAQUO_QUPATH_DIR={app}")
    else:
        print("#  REM Windows CMD")
        print(f'#  C:\\> set PAQUO_QUPATH_DIR="{app}"')
        print("#  # Windows PowerShell")
        print(f'#  PS C:\\> $env:PAQUO_QUPATH_DIR="{app}"')
    print("#\n# use via .paquo.toml config file:")
    print(f'#  qupath_dir="{app}"')
    print(app)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



"""example showing how to create a project with annotations"""
from pathlib import Path
from paquo.projects import QuPathProject
from paquo.images import QuPathImageType
from shapely.geometry import Point, Polygon, LineString

EXAMPLE_PROJECT = Path(__file__).parent.absolute() / "projects" / "example_02_project"
EXAMPLE_IMAGE = Path(__file__).parent.absolute() / "images" / "image_1.svs"

ANNOTATIONS = {
    'Annotation 1': Point(500, 500),
    'Annotation 2': Polygon.from_bounds(510, 400, 610, 600),
    'Some Other Name': LineString([[400, 400], [450, 450], [400, 425]])
}

# create a the new project
with QuPathProject(EXAMPLE_PROJECT, mode='x') as qp:
    print("created", qp.name)

    # add a new image:
    entry = qp.add_image(EXAMPLE_IMAGE, image_type=QuPathImageType.BRIGHTFIELD_H_E)

    for name, roi in ANNOTATIONS.items():
        # add the annotations without a class set
        annotation = entry.hierarchy.add_annotation(roi=roi)
        annotation.name = name

    print(f"done. Please look at {qp.name} in QuPath.")



"""example showing how to create a empty project with classes"""
from pathlib import Path
from paquo.projects import QuPathProject
from paquo.classes import QuPathPathClass

EXAMPLE_PROJECT = Path(__file__).parent.absolute() / "projects" / "example_03_project"

MY_CLASSES_AND_COLORS = [
    ("My Class 1", "#ff0000"),
    ("Some Other Class", "#0000ff"),
    ("Nothing*", "#00ff00"),
]

# create a the new project
with QuPathProject(EXAMPLE_PROJECT, mode='x') as qp:
    print("created", qp.name)

    new_classes = []
    for class_name, class_color in MY_CLASSES_AND_COLORS:
        new_classes.append(
            QuPathPathClass(name=class_name, color=class_color)
        )

    # setting QuPathProject.path_class always replaces all classes
    qp.path_classes = new_classes
    print("project classes:")
    for path_class in qp.path_classes:
        print(">", f"'{path_class.name}'", path_class.color.to_hex())

    print(f"done. Please look at {qp.name} in QuPath.")



import atexit
import logging
import re
from contextlib import AbstractContextManager
from contextlib import ContextDecorator
from contextlib import ExitStack
from contextlib import contextmanager
from contextlib import suppress
from typing import Iterable
from typing import List
from typing import Tuple

from paquo.java import ByteArrayOutputStream
from paquo.java import LogManager
from paquo.java import PrintStream
from paquo.java import StandardCharsets
from paquo.java import System

__all__ = ['get_logger', 'redirect']

# === qupath logging setup ============================================

# by default, silence the qupath java logging if nothing is configured
# https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
logging.getLogger('qupath').addHandler(logging.NullHandler())

# set the java log level to DEBUG
# and filter in the 'qupath' logger on the python side
if LogManager:
    LogManager.setDebug()


@atexit.register
def _java_log_cleanup():
    # NOTE: qupath's ImageServer Threads provide debug output that's
    #   only emitted once the interpreter is cleaned up. Since at
    #   that point we've already left the contextmanager for log
    #   redirection, we'll just switch the qupath loglevel to Warn
    #   and emit a debug message on the python side.
    if LogManager:
        LogManager.setWarn()
    get_logger('qupath').debug("switching java logger to warn for silent thread cleanup")


# --- logging methods -------------------------------------------------

def get_logger(name):
    """return a logger instance"""
    return logging.getLogger(name)


class _JavaLoggingBase(AbstractContextManager):
    """reentrant logging abstraction for redirecting JVM output to python

    will currently flush logging output on exit
    """
    java_default = None  # REQUIRED IN SUBCLASSES
    java_setter = None  # REQUIRED IN SUBCLASSES
    _count = 0
    _java_buffer = None
    _logger = logging.getLogger("qupath")
    _java_log_entry_match = re.compile(r"""^
        (?P<timestamp>[0-9:.]+)
        [ ]\[(?P<logger>[^]]+)]
        [ ]\[(?P<level>[^]]+)]
        [ ](?P<origin>[^ ]+)
        [ ]-[ ](?P<msg>.*)
        $
    """, re.VERBOSE).match

    @contextmanager
    def _stop_redirection_on_error(self):
        # reuse stdio reset in case __enter__ crashes
        with ExitStack() as stack:
            stack.push(self)
            yield
            # we did not crash!
            stack.pop_all()

    def __enter__(self):
        """increase reference count and redirect stdio"""
        self._count += 1
        with self._stop_redirection_on_error():
            if self._java_buffer is None:
                java_buffer = ByteArrayOutputStream()
                ps = PrintStream(
                    java_buffer,
                    True,
                    StandardCharsets.UTF_8.name()
                )
                # note: these two lines should be made atomic
                self.java_setter(ps)
                self._java_buffer = java_buffer
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """decrease reference count and cleanup if 0"""
        self._count = max(0, self._count - 1)
        self.flush_logs()  # flush logs no matter what
        if not self._count:
            # note: these two lines should be made atomic
            self.java_setter(self.java_default)
            self._java_buffer = None

    def flush_logs(self):
        """flush the java buffer to the Logger"""
        # extract the buffer and clear it
        try:
            output = str(self._java_buffer.toString())
        except AttributeError:  # pragma: no cover
            return
        with suppress(AttributeError):
            self._java_buffer.reset()
        # assume JVM console output is one line per msg
        for (origin, level), entry in self.iter_logs(output):
            if "WARN" in level:
                self._logger.warning("[%s] [%s] - %s", origin, level, entry)
            elif "ERR" in level:
                self._logger.error("[%s] [%s] - %s", origin, level, entry)  # pragma: no cover
            elif "DEBUG" in level:
                self._logger.debug("[%s] [%s] - %s", origin, level, entry)
            else:
                self._logger.info("[%s] [%s] - %s", origin, level, entry)

    def iter_logs(self, output: str) -> Iterable[Tuple[Tuple[str, str], str]]:
        """iterate the individual log messages"""
        entry: List[str] = []
        info = ('NONE', 'NONE')
        for line in output.splitlines(keepends=True):
            if not line.strip():
                continue  # pragma: no cover
            m = self.__class__._java_log_entry_match(line)
            if m:
                if entry:
                    yield info, "".join(entry).rstrip()
                    entry.clear()
                info = (
                    m.group('origin'),
                    m.group('level').strip().upper(),
                )
                entry.append(m.group('msg'))
            else:
                entry.append(line)  # pragma: no cover
        if entry:
            yield info, "".join(entry).rstrip()


class _JavaLoggingStdout(_JavaLoggingBase):
    java_default = System.out
    java_setter = System.setOut


class _JavaLoggingStderr(_JavaLoggingBase):
    java_default = System.err
    java_setter = System.setErr


# noinspection PyPep8Naming
class redirect(ExitStack, ContextDecorator):
    """convenient contextdecorator for redirecting JVM output"""

    def __init__(self, stdout=True, stderr=True):
        super().__init__()
        self._stdout = _JavaLoggingStdout if stdout else None
        self._stderr = _JavaLoggingStderr if stderr else None

    def __enter__(self):
        super().__enter__()
        if self._stderr:
            self.enter_context(self._stderr())
        if self._stdout:
            self.enter_context(self._stdout())
        return self



import json
import lzma
import os
import platform
import re
import shutil
import ssl
import subprocess
import sys
import tarfile
import tempfile
import warnings
import zipfile
from datetime import datetime
from functools import cached_property as _cached_property
from functools import partial
from functools import total_ordering
from pathlib import Path
from urllib.parse import urlsplit
from urllib.request import urlopen
from warnings import warn

from packaging.version import Version

__all__ = [
    'QuPathVersion',
    'cached_property',
    'make_backup_filename',
    'load_json_from_path'
]


# noinspection PyPep8Naming
class cached_property(_cached_property):
    def __set__(self, obj, value):
        raise AttributeError(f"readonly attribute {self.attrname}")


@total_ordering
class QuPathVersion:
    """Handle the QuPath version strings"""
    def __init__(self, version: str) -> None:
        self.origin = ver_str = str(version).strip()
        # to not having to rely on packaging.version.LegacyVersion
        # we replace the milestone versioning with -devNN
        ver_str = re.sub(r"(?:-)m(?P<num>[0-9]+)", r"dev\g<num>", ver_str, count=1)
        # to allow running snapshot versions we replace the suffix
        # with a local version
        ver_str = ver_str.replace("-SNAPSHOT", "+snapshot")
        self.version = Version(ver_str)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.origin!r})"

    def __str__(self) -> str:
        return self.origin

    def __eq__(self, other) -> bool:
        if isinstance(other, QuPathVersion):
            return self.version.__eq__(other.version)
        else:
            return self.version.__eq__(other)

    def __lt__(self, other) -> bool:
        if isinstance(other, QuPathVersion):
            return self.version.__lt__(other.version)
        else:
            return self.version.__lt__(other)


def make_backup_filename(path, name, suffix='backup'):
    path = Path(path)
    if not path.is_dir():
        raise ValueError("requires a directory")  # pragma: no cover
    now = datetime.now().strftime('%Y%m%d-%H%M%S')
    backup = path / f"{name}-{now}.{suffix}"
    if backup.is_file():
        raise RuntimeError(f"the file {backup} should not exist!")  # pragma: no cover
    return backup


def load_json_from_path(path):
    path = Path(path)
    if not path.is_file():
        raise FileNotFoundError(str(path))  # pragma: no cover

    if path.name.endswith(".geojson.xz"):
        ctx = partial(lzma.open, path, 'rt')
    elif path.name.endswith(('.geojson', '.json')):
        ctx = partial(path.open, 'r')
    else:
        raise NotImplementedError(f"unsupported file format '{path}'")

    with ctx() as fobj:
        data = json.load(fobj)

    if isinstance(data, dict):
        return data
    elif isinstance(data, list):
        return {'annotations': data}
    else:
        raise ValueError("expected dict or list of annotations")


def download_qupath(
    version,
    path,
    *,
    system=None,
    callback=(lambda chunk_iter, name: chunk_iter),
    ssl_verify=True
):
    """download qupath from github"""
    if system is None:
        system = platform.system()

    if system == "Linux":
        _sys = "Linux"
        ext = "tar.xz"
    elif system == "Darwin":
        _sys = "Mac"
        ext = "pkg"
    elif system == "Windows":
        _sys = "Windows"
        ext = "zip"
    else:
        raise ValueError(f"unsupported platform.system() == {system!r}")

    if not version.startswith("v"):
        version = f"v{version}"

    if Version(version) > Version("0.4.4"):
        if system == "Darwin":
            if platform.machine() == "arm64":
                _sys = "Mac-arm64"
            else:
                _sys = "Mac-x64"
        name = f"QuPath-{version}-{_sys}"
    elif Version(version) > Version("0.3.2"):
        if system == "Darwin":
            if platform.machine() == "arm64":
                _sys = "Mac-arm64"
            else:
                _sys = "Mac"
        name = f"QuPath-{version[1:]}-{_sys}"
    else:
        if "rc" not in version:
            name = f"QuPath-{version[1:]}-{_sys}"
        else:
            name = f"QuPath-{version[1:]}"

    url = f"https://github.com/qupath/qupath/releases/download/{version}/{name}.{ext}"

    chunk_size = 10 * 1024 * 1024

    fn = os.path.basename(urlsplit(url).path)
    out_fn = os.path.join(path, fn)
    if os.path.exists(out_fn):
        return out_fn

    if ssl_verify:
        _ctx = None
    else:
        warn(f"DISABLING SSL VERIFICATION FOR: {url}", stacklevel=2)
        _ctx = ssl.create_default_context()
        _ctx.check_hostname = False
        _ctx.verify_mode = ssl.CERT_NONE

    try:
        with open(out_fn, mode="wb") as tmp, urlopen(url, context=_ctx) as f:  # nosec B310
            for chunk in callback(iter(lambda: f.read(chunk_size), b""), name=url):
                tmp.write(chunk)
    except Exception:
        print("# error requesting:", url, file=sys.stderr)
        try:
            os.unlink(out_fn)
        except OSError:
            pass
        raise
    else:
        return out_fn


def extract_qupath(file, destination, system=None):
    """extract downloaded QuPath file to a destination"""
    fn = os.path.basename(file)

    # normalize QuPath App dirname
    m = re.match(
        r"QuPath-v?(?P<version>[0-9]+[.][0-9]+[.][0-9]+(-rc[0-9]+|-m[0-9]+)?)",
        fn,
    )

    if system is None:
        system = platform.system()

    if system in {"Linux", "Windows"}:
        app_dir = f"QuPath-{m.group('version')}"
    elif system == "Darwin":
        app_dir = f"QuPath-{m.group('version')}.app"
    else:
        raise ValueError(f"unsupported platform.system() == {system!r}")

    destination = os.path.abspath(os.path.expanduser(destination))
    if not os.path.isdir(destination):
        raise ValueError(f"destination: {destination!r} is not a directory")
    qp_dst = os.path.join(destination, app_dir)
    if os.path.isdir(qp_dst):
        warnings.warn(
            f"Skipping! Output directory already exists: {qp_dst!r}",
            stacklevel=2,
        )
        return qp_dst

    if system == "Linux":
        if not os.fspath(file).endswith(".tar.xz"):
            raise ValueError("file does not end with `.tar.xz`")
        with tempfile.TemporaryDirectory() as tmp_dir:
            with tarfile.open(file, mode="r:xz") as tf:
                tf.extractall(tmp_dir)  # nosec: B202
                for name in os.listdir(tmp_dir):
                    pth = os.path.join(tmp_dir, name)
                    if name.startswith("QuPath") and os.path.isdir(pth):
                        break
                else:
                    raise RuntimeError("no qupath extracted?")
            extract_dir = os.path.join(tmp_dir, name)
            if os.path.isdir(os.path.join(extract_dir, "QuPath")):
                # in some cases there is a nested QuPath directory
                extract_dir = os.path.join(extract_dir, "QuPath")
            shutil.move(extract_dir, qp_dst)
        return qp_dst

    elif system == "Darwin":
        if not Path(file).suffix == ".pkg":
            raise ValueError("file does not end with `.pkg`")
        if shutil.which("7z") is None:
            raise RuntimeError("7z is required, run: `brew install p7zip`")

        with tempfile.TemporaryDirectory() as tmp_dir:
            subprocess.run(["7z", "x", os.path.abspath(file)], cwd=tmp_dir, capture_output=True)
            subprocess.run(["7z", "x", "Payload~"], cwd=tmp_dir, capture_output=True)
            for name in os.listdir(tmp_dir):
                if name.startswith("QuPath") and name.endswith(".app"):
                    break
            else:
                raise RuntimeError("no qupath extracted?")
            shutil.move(os.path.join(tmp_dir, name), qp_dst)
        return qp_dst

    elif system == "Windows":
        if not Path(file).suffix == ".zip":
            raise ValueError("file does not end with `.zip`")

        with tempfile.TemporaryDirectory() as tmp_dir:
            with zipfile.ZipFile(file, mode="r") as zf:
                zf.extractall(tmp_dir)  # nosec: B202
                for name in os.listdir(tmp_dir):
                    pth = os.path.join(tmp_dir, name)
                    if name.startswith("QuPath") and os.path.isdir(pth):
                        break
                    if name.startswith("QuPath") and name.endswith(".exe") and os.path.isfile(pth):
                        pth = tmp_dir
                        break
                else:
                    raise RuntimeError("no qupath extracted?")
            shutil.move(pth, qp_dst)
        return qp_dst

    else:
        raise ValueError(f"unsupported platform.system() == {system!r}")



import pytest
import shapely.geometry
from shapely import Polygon

from paquo._utils import QuPathVersion
from paquo.classes import QuPathPathClass
from paquo.pathobjects import QuPathPathAnnotationObject, QuPathPathCellObject
from paquo.projects import QuPathProject


def test_qupath_to_shapely_conversion():
    from paquo.java import ROIs
    from paquo.pathobjects import _qupath_roi_to_shapely_geometry
    from shapely.geometry import Point

    roi = ROIs.createPointsROI(1.0, 2.0, None)
    point = _qupath_roi_to_shapely_geometry(roi)

    assert isinstance(point, Point)
    assert point.x == 1.0 and point.y == 2.0


def test_shapely_to_qupath_conversion():
    from paquo.pathobjects import _shapely_geometry_to_qupath_roi
    from shapely.geometry import Point

    point = Point(1.0, 2.0)
    roi = _shapely_geometry_to_qupath_roi(point)

    assert roi.getRoiName() == "Points"


@pytest.fixture(
    scope="function",
    params=[
        shapely.geometry.Polygon.from_bounds(10, 20, 100, 200),
        shapely.geometry.Point(1, 2),
        shapely.geometry.LineString([[1, 1], [2, 2]]),
        shapely.geometry.MultiPolygon(
            [shapely.geometry.Polygon.from_bounds(10, 20, 100, 200),
             shapely.geometry.Polygon.from_bounds(110, 20, 200, 200)]
        )
    ],
    ids=[
        "Polygon",
        "Point",
        "LineString",
        "MultiPolygon"
    ],
)
def path_annotation(request):
    """parameterized fixture for different Annotation Objects"""
    roi = request.param
    path_class = QuPathPathClass("myclass")

    ao = QuPathPathAnnotationObject.from_shapely(roi, path_class)

    yield ao


def test_geojson_serialization(path_annotation, qupath_version):
    geo_json = path_annotation.to_geojson()

    assert geo_json["type"] == "Feature"

    # bad practice
    if qupath_version <= QuPathVersion("0.2.3"):
        assert geo_json["id"] == "PathAnnotationObject"
    elif qupath_version < QuPathVersion("0.4.0"):
        assert geo_json["properties"]["object_type"] == "annotation"
    else:
        assert geo_json["properties"]["objectType"] == "annotation"

    assert "geometry" in geo_json
    geom = geo_json["geometry"]

    p = shapely.geometry.shape(geom)
    assert p.equals(path_annotation.roi)

    assert "properties" in geo_json
    prop = geo_json["properties"]

    if path_annotation.locked:
        assert prop["isLocked"] is True
    else:
        if qupath_version < QuPathVersion("0.4.0"):
            assert prop["isLocked"] is False
        else:
            assert "isLocked" not in prop

    # bad practice
    if qupath_version <= QuPathVersion("0.2.3"):
        assert prop["measurements"] == path_annotation.measurements.to_records()
    else:
        measurements = path_annotation.measurements.to_records()
        if not measurements:
            assert "measurements" not in prop
        else:
            assert prop["measurements"] == measurements

    assert prop["classification"]["name"] == path_annotation.path_class.name
    if qupath_version < QuPathVersion("0.4.0"):
        assert prop["classification"]["colorRGB"] == path_annotation.path_class.color.to_java_rgba()
    else:
        assert tuple(prop["classification"]["color"]) == path_annotation.path_class.color.to_rgb()


def test_annotation_object():
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        QuPathPathAnnotationObject.from_shapely(roi=123)

    ao = QuPathPathAnnotationObject.from_shapely(
        shapely.geometry.Point(1, 1),
        path_class=QuPathPathClass('myclass'),
        measurements={'measurement1': 1.23},
        path_class_probability=0.5,
    )

    assert ao.path_class.name == "myclass"
    assert ao.path_class_probability == 0.5
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        ao.update_path_class(123)
    ao.update_path_class(None)
    assert ao.path_class is None
    assert ao.is_editable
    assert not ao.locked
    ao.locked = True
    assert ao.locked
    assert ao.level == 0

    # name
    assert ao.name is None
    ao.name = "my annotation"
    assert ao.name == "my annotation"
    ao.name = None

    # description
    assert ao.description is None
    ao.description = "my description"
    assert ao.description == "my description"
    with pytest.raises(TypeError):
        ao.description = None

    # roi
    pt = shapely.geometry.Point(2, 2)
    ao.update_roi(pt)
    assert ao.roi == pt

    # repr
    ao.name = "abc"
    ao.update_path_class(QuPathPathClass('myclass'))
    assert repr(ao)


def test_pathcellobject():

    roi = shapely.geometry.Polygon.from_bounds(10, 20, 100, 200)
    nucleus_roi = shapely.geometry.Polygon.from_bounds(14, 26, 90, 190)

    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        QuPathPathCellObject.from_shapely(roi=123, nucleus_roi=123)
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        QuPathPathCellObject.from_shapely(roi=roi)

    cell = QuPathPathCellObject.from_shapely(
        roi,
        path_class=QuPathPathClass('myclass'),
        nucleus_roi=nucleus_roi,
    )

    assert shapely.equals(cell.roi, roi)
    assert shapely.equals(cell.nucleus_roi, nucleus_roi)


def test_measurements():
    ao = QuPathPathAnnotationObject.from_shapely(
        shapely.geometry.Point(1, 1),
        path_class=QuPathPathClass('myclass'),
        measurements={'measurement1': 1.23},
        path_class_probability=0.5,
    )

    assert 123 not in ao.measurements
    assert "measurement1" in ao.measurements
    assert len(ao.measurements) == 1
    assert repr(ao.measurements)
    assert str(ao.measurements)

    # allow index access (note this depends on order of insertion)
    assert ao.measurements[0] == ao.measurements['measurement1']

    with pytest.raises(KeyError):
        _ = ao.measurements[None]

    with pytest.raises(KeyError):
        _ = ao.measurements["this-key-does-not-exist"]

    with pytest.raises(KeyError):
        del ao.measurements[None]

    del ao.measurements['measurement1']
    ao.measurements.clear()


@pytest.fixture(scope="function")
def project_with_detections(tmp_path, svs_small):
    pth = tmp_path / "paquo-project"
    with QuPathProject(pth, mode='x') as proj:
        entry = proj.add_image(svs_small)
        entry.hierarchy.add_detection(
            roi=Polygon.from_bounds(20, 20, 100, 100),
        )
        assert len(entry.hierarchy.detections) == 1
    yield pth


def test_saving_measurements(project_with_detections):
    with QuPathProject(project_with_detections, mode="a") as proj:
        det = proj.images[0].hierarchy.detections[0]
        assert len(det.measurements) == 0
        det.measurements["something"] = 1.0
        assert dict(det.measurements) == {"something": 1.0}

    with QuPathProject(project_with_detections, mode="r") as proj:
        det = proj.images[0].hierarchy.detections[0]
        assert det.measurements["something"] == 1.0



import pytest

from paquo.colors import QuPathColor


def test_incorrect_color():
    # NOTE THIS DOESNT CRASH BY DESIGN CURRENTLY!
    c = QuPathColor(300, 300, 300)
    assert not c.is_valid()
    c = QuPathColor(30, 30, 30)
    assert c.is_valid()


def test_color_roundtrip():
    c0 = (64, 64, 64)
    j0 = QuPathColor(*c0).to_java_rgb()
    c1 = QuPathColor.from_java_rgb(j0).to_rgb()
    assert c0 == c1


def test_color_mpl_rgba():
    assert QuPathColor(255, 255, 255).to_mpl_rgba() == (1, 1, 1, 1)


def test_repr():
    repr(QuPathColor(0, 0, 0, 255))
    repr(QuPathColor(0, 0, 0, 0))


def test_color_from_color():
    c = QuPathColor(0, 0, 0, 0)
    assert QuPathColor.from_any(c) == c


def test_from_any_error():
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        QuPathColor.from_any(object())


def test_alpha_roundtrip():
    c0 = QuPathColor(1, 2, 3, 255)
    j0 = c0.to_java_rgba()
    c1 = QuPathColor.from_java_rgba(j0)

    assert c0 == c1
    assert c1.alpha == 255 == c0.alpha


def test_rgba_via_java_rgb():
    c0 = QuPathColor(1, 2, 3, 4)  # alpha will be lost
    j0 = c0.to_java_rgb()
    c1 = QuPathColor.from_java_rgba(j0)

    assert c0.red == c1.red
    assert c0.green == c1.green
    assert c0.blue == c1.blue
    assert c0.alpha == 4 and c1.alpha == 255


def test_rgb_via_java_rgba():
    c0 = QuPathColor(1, 2, 3, 4)
    j0 = c0.to_java_rgba()
    c1 = QuPathColor.from_java_rgb(j0)

    assert c0.red == c1.red
    assert c0.green == c1.green
    assert c0.blue == c1.blue
    assert c0.alpha == 4 and c1.alpha == 255


def test_from_java_typecheck():
    QuPathColor.from_java_rgb(0)
    QuPathColor.from_java_rgba(0)
    with pytest.raises(TypeError):
        QuPathColor.from_java_rgb(None)
    with pytest.raises(TypeError):
        QuPathColor.from_java_rgba(None)


def test_hexcolor():
    c0 = "#ff00ff"
    qc = QuPathColor.from_hex(c0)
    assert qc.to_rgb() == (255, 0, 255)
    assert c0 == qc.to_hex()

    with pytest.raises(ValueError):
        QuPathColor.from_hex("abc")

    assert qc == QuPathColor.from_any(c0)



import platform
import sys
from argparse import ArgumentTypeError
from collections import defaultdict
from functools import partial
from pathlib import Path

# -- argparse improvements ---------------------------------------------


def subcommand(*arguments, parent):
    """decorator helper for commandline"""
    def decorator(func):
        fn = func.__name__.rstrip('_')
        started_via_m = Path(sys.argv[0]).name == "__main__.py"
        subparser = parent.add_parser(
            name=fn,
            prog=f"python -m paquo {fn}" if started_via_m else f"paquo {fn}",
            help=func.__doc__,
        )
        for args, kwargs in arguments:
            subparser.add_argument(*args, **kwargs)
        subparser.set_defaults(cmd_func=partial(func, subparser=subparser))
        return func
    return decorator


def argument(*args, **kwargs):
    """argument helper for subcommand"""
    return args, kwargs


class DirectoryType:
    """Directory parsing for argparse"""
    def __call__(self, cmd_input: str):
        p = Path(cmd_input)
        if p.is_dir():
            return p
        raise ArgumentTypeError(f"'{cmd_input}' is not a directory")


# -- config related commands -------------------------------------------

def config_print_settings():
    """print the current configuration"""
    from paquo import settings
    from paquo._config import to_toml

    print("# current paquo configuration")
    print("# ===========================")
    print("# format: TOML")
    print(to_toml(settings))


def config_print_defaults():
    """print the default paquo configuration"""
    if sys.version_info >= (3, 9):
        from importlib.resources import files

        def read_text(package, resource, encoding):
            return files(package).joinpath(resource).read_text(encoding=encoding)
    else:
        from importlib.resources import read_text

    from paquo._config import settings

    output = read_text(
        "paquo",
        ".paquo.defaults.toml",
        encoding=settings.ENCODING_FOR_DYNACONF
    )
    print(output)


# -- create related commands -------------------------------------------

def list_project(path):
    """print information about the project"""
    from paquo.projects import QuPathProject

    with QuPathProject(path, mode='r') as qp:
        print("Project:", qp.name)
        print("Classes: #", len(qp.path_classes), sep='')
        for path_class in qp.path_classes:
            color = path_class.color
            color_str = f" [{color.to_hex()}]" if color else ""
            print(f"- {path_class.id}{color_str}")

        _md_keys = set()
        _md_value_len = defaultdict(int)
        print("Images: #", len(qp.images), sep='')
        for idx, image in enumerate(qp.images):
            print(f"- [{idx}] {image.image_name}")
            for k, v in image.metadata.items():
                _md_keys.add(k)
                _md_value_len[k] = max(_md_value_len[k], len(v), len(k))

        print("Project Metadata: #", len(_md_keys), sep='')
        if len(_md_keys):
            _md_keys = list(sorted(_md_keys))
            hdr = ["#    "]
            for _key in _md_keys:
                hdr.append(f"{_key:<{_md_value_len[_key]}}")
            print(" ".join(hdr))
            for idx, image in enumerate(qp.images):
                line = [f"- [{idx}]"]
                for _key in _md_keys:
                    val = image.metadata.get(_key, '')
                    line.append(f"{val:<{_md_value_len[_key]}}")
                print(" ".join(line))


# -- create related commands -------------------------------------------

def create_project(project_path, class_names_colors, images,
                   annotations_json_func=None,
                   remove_default_classes=False, force_write=False):
    """create a qupath project"""
    from paquo._logging import get_logger
    from paquo._utils import load_json_from_path
    from paquo.classes import QuPathPathClass
    from paquo.images import QuPathImageType
    from paquo.projects import QuPathProject

    _logger = get_logger(__name__)

    mode = 'x' if not force_write else 'w'

    if annotations_json_func is not None and not callable(annotations_json_func):
        raise ValueError("annotations_json_func should be callable")

    # noinspection PyTypeChecker
    with QuPathProject(project_path, mode=mode) as qp:
        if remove_default_classes:
            qp.path_classes = ()
        path_classes = list(qp.path_classes)
        for name, color in class_names_colors:
            path_classes.append(
                QuPathPathClass(name, color=color)
            )
        qp.path_classes = path_classes

        for image in images:
            qp_image = qp.add_image(image, image_type=QuPathImageType.BRIGHTFIELD_H_E)

            if annotations_json_func:
                annotations_jsons = annotations_json_func(Path(image).name)
                for annotations_json in annotations_jsons:
                    _logger.info(f"loading '{annotations_json}'")
                    geojson = load_json_from_path(annotations_json)
                    qp_image.hierarchy.load_geojson(geojson["annotations"])

        name = qp.name
    return name


# -- export related commands -------------------------------------------

def export_annotations(path, image_idx, pretty=False):
    """print annotations as geojson"""
    import pprint

    from paquo.projects import QuPathProject

    with QuPathProject(path, mode='r') as qp:
        image = qp.images[image_idx]
        data = image.hierarchy.to_geojson()
        if pretty:
            pprint.pprint(data)
        else:
            print(data)


# -- open related commands ------------------------------------------

def open_qupath(project_path):
    """launch qupath with the provided project"""
    import subprocess
    from contextlib import ExitStack
    from contextlib import contextmanager
    from tempfile import TemporaryDirectory
    from zipfile import ZipFile

    from paquo._config import settings
    from paquo._config import to_kwargs
    from paquo.jpype_backend import find_qupath

    # retrieve the path of the qupath executable
    app_dir, _, _, _ = find_qupath(**to_kwargs(settings))
    system = platform.system()
    if system == "Linux":
        qupath, = Path(app_dir).parent.parent.joinpath("bin").glob("QuPath*")

    elif system == "Darwin":
        qupath, = Path(app_dir).parent.joinpath("MacOS").glob("QuPath*")

    elif system == "Windows":
        _qp_exes = list(Path(app_dir).parent.glob("QuPath*.exe"))
        assert len(_qp_exes) == 2, f"this should have returned two paths, got {_qp_exes}"
        qupath, = (qp for qp in _qp_exes if "console" in qp.stem)

    else:
        raise ValueError(f"Unknown platform {system}")

    @contextmanager
    def prepare_dir(path: Path):
        # unzip a qpzip file to a temporary directory
        if path.is_file() and path.suffix == ".qpzip":
            with ExitStack() as stack:
                tmp_path = stack.enter_context(
                    TemporaryDirectory(prefix="pado", suffix="qpzip")
                )
                with ZipFile(path) as qpzip:
                    qpzip.extractall(tmp_path)
                print(f"Extracted qpzip to {tmp_path}")
                yield Path(tmp_path)
                # dont clean up tmp if no errors were raised
                stack.pop_all()
        else:
            yield path

    with prepare_dir(Path(project_path)) as p:
        if p.is_dir():
            p /= "project.qpproj"
        if not (p.is_file() and p.suffix == ".qpproj"):
            raise ValueError(f"Not a qupath project: '{p}'")

        subprocess.run([qupath, '-p', p.resolve()])


# -- quzip related commands -------------------------------------------

def qpzip_project(project_path):
    """create a zip archive of a qupath project that can be used with `paquo open`"""
    import shutil
    from tempfile import TemporaryDirectory

    p = Path(project_path).expanduser().absolute()
    if p.is_dir():
        p /= "project.qpproj"
    if not (p.is_file() and p.suffix == ".qpproj"):
        raise ValueError(f"Not a qupath project: '{p}'")

    project_path = p.parent
    with TemporaryDirectory() as tmpdir:
        tmp_base_name = Path(tmpdir) / project_path.name
        tmp_zip = shutil.make_archive(tmp_base_name, "zip", root_dir=project_path, base_dir=".")
        qpzip = shutil.move(tmp_zip, project_path.with_suffix(".qpzip"))
    print(f"created: {qpzip}")
